@@ -1,9 +1,15 @@
 """Classes and supporting functions to manipulate ordering columns and extract
 keyset markers from query results."""
+from __future__ import annotations
+
+from abc import ABC, abstractmethod
 from copy import copy
+from typing import List, Optional
 from warnings import warn
 
 import sqlalchemy
+import sqlalchemy.exc
+import sqlalchemy.orm.exc
 from sqlalchemy import asc, column
 from sqlalchemy.orm import Bundle, Mapper, class_mapper
 from sqlalchemy.orm.attributes import QueryableAttribute
@@ -27,24 +33,6 @@ _WRAPPING_OVERFLOW = (
 )
 
 
-def parse_ob_clause(selectable):
-    """Parse the ORDER BY clause of a selectable into a list of :class:`OC` instances."""
-
-    def _flatten(cl):
-        if isinstance(cl, ClauseList):
-            for subclause in cl.clauses:
-                for x in _flatten(subclause):
-                    yield x
-        elif isinstance(cl, (tuple, list)):
-            for xs in cl:
-                for x in _flatten(xs):
-                    yield x
-        else:
-            yield cl
-
-    return [OC(c) for c in _flatten(order_by_clauses(selectable))]
-
-
 def _warn_if_nullable(x):
     try:
         if x.nullable or x.property.columns[0].nullable:
@@ -94,7 +82,7 @@ class OC:
         return str(self).split()[0]
 
     @property
-    def element(self):
+    def element(self) -> ColumnElement:
         """The ordering column/SQL expression with ordering modifier removed."""
         return _remove_order_direction(self.uo)
 
@@ -136,7 +124,7 @@ class OC:
         # If this OC is a column with a custom type, apply the custom
         # preprocessing to the comparsion value:
         try:
-            value = compval.type.bind_processor(dialect)(value)
+            value = compval.type.bind_processor(dialect)(value)  # type: ignore
         except (TypeError, AttributeError):
             pass
         if self.is_ascending:
@@ -151,7 +139,25 @@ class OC:
         return "<OC: {}>".format(str(self))
 
 
-def strip_labels(el):
+def parse_ob_clause(selectable) -> List[OC]:
+    """Parse the ORDER BY clause of a selectable into a list of :class:`OC` instances."""
+
+    def _flatten(cl):
+        if isinstance(cl, ClauseList):
+            for subclause in cl.clauses:
+                for x in _flatten(subclause):
+                    yield x
+        elif isinstance(cl, (tuple, list)):
+            for xs in cl:
+                for x in _flatten(xs):
+                    yield x
+        else:
+            yield cl
+
+    return [OC(c) for c in _flatten(order_by_clauses(selectable))]
+
+
+def strip_labels(el: ColumnElement) -> ColumnElement:
     """Remove labels from a
     :class:`sqlalchemy.sql.expression.ColumnElement`."""
     while isinstance(el, _LABELLED):
@@ -182,7 +188,7 @@ def _get_order_direction(x):
     raise Exception(_WRAPPING_OVERFLOW)  # pragma: no cover
 
 
-def _reverse_order_direction(ce):
+def _reverse_order_direction(ce: ColumnElement):
     """
     Given a :class:`sqlalchemy.sql.expression.ColumnElement`, return a copy
     with its ordering direction (ASC or DESC) reversed (if it has one).
@@ -208,7 +214,7 @@ def _reverse_order_direction(ce):
     raise Exception(_WRAPPING_OVERFLOW)  # pragma: no cover
 
 
-def _remove_order_direction(ce):
+def _remove_order_direction(ce: ColumnElement) -> ColumnElement:
     """
     Given a :class:`sqlalchemy.sql.expression.ColumnElement`, return a copy
     with its ordering modifiers (ASC/DESC, NULLS FIRST/LAST) removed (if it has
@@ -248,7 +254,7 @@ def _remove_order_direction(ce):
     raise Exception(_WRAPPING_OVERFLOW)  # pragma: no cover
 
 
-class MappedOrderColumn:
+class MappedOrderColumn(ABC):
     """An ordering column in the context of a particular query/select.
 
     This wraps an :class:`OC` with one extra piece of information: how to
@@ -256,16 +262,19 @@ class MappedOrderColumn:
     this requires adding extra entities to the query; in this case,
     ``extra_column`` will be set."""
 
-    def __init__(self, oc):
+    oc: OC
+    extra_column: Optional[ColumnElement]
+    """An extra SQLAlchemy ORM entity that this ordering column needs to
+    add to its query in order to retrieve its value at each row. If no
+    extra data is required, the value of this property will be ``None``."""
+
+    def __init__(self, oc: OC):
         self.oc = oc
         self.extra_column = None
-        """An extra SQLAlchemy ORM entity that this ordering column needs to
-        add to its query in order to retrieve its value at each row. If no
-        extra data is required, the value of this property will be ``None``."""
 
+    @abstractmethod
     def get_from_row(self, internal_row):
         """Extract the value of this ordering column from a result row."""
-        raise NotImplementedError  # pragma: no cover
 
     @property
     def ob_clause(self):
@@ -320,6 +329,7 @@ class AppendedColumn(MappedOrderColumn):
     original query."""
 
     _counter = 0
+    extra_column: ColumnElement
 
     def __init__(self, oc, name=None):
         super().__init__(oc)
@@ -405,7 +415,7 @@ def derive_order_key(ocol, desc, index):
         pass
 
 
-def find_order_key(ocol, column_descriptions):
+def find_order_key(ocol: OC, column_descriptions) -> MappedOrderColumn:
     """Return a :class:`MappedOrderColumn` describing how to populate the
     ordering column `ocol` from a query returning columns described by
     `column_descriptions`.
