@@ -1,17 +1,40 @@
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Union
 import sqlalchemy
 from packaging import version
+from sqlalchemy.engine import Connection, Engine
+from sqlalchemy.orm import Session
 
 
 SQLA_VERSION = version.parse(sqlalchemy.__version__)
 
 
-def get_bind(q, s):
-    try:
-        # session
+try:
+    from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncSession
+except ImportError:
+    if not TYPE_CHECKING:
+        class AsyncConnection:
+            pass
+
+        class AsyncEngine:
+            pass
+
+        class AsyncSession:
+            pass
+
+
+def get_bind(
+    q, s: Union[Engine, Connection, Session, AsyncEngine, AsyncConnection, AsyncSession]
+) -> Union[Engine, Connection]:
+    if isinstance(s, (Session, AsyncSession)):
         return s.get_bind(clause=getattr(q, "statement", q))
-    except Exception:
-        # connection/engine
+    elif isinstance(s, (Engine, Connection)):
         return s
+    elif isinstance(s, (AsyncEngine, AsyncConnection)):
+        return s.sync_engine
+    else:
+        raise ValueError(f"{s} is not a (sync/async) Engine, Connection or Session.")
 
 
 if SQLA_VERSION < version.parse("1.4.0b1"):
@@ -24,6 +47,7 @@ if SQLA_VERSION < version.parse("1.4.0b1"):
         orm_query_keys,
         orm_result_type,
         orm_to_selectable,
+        Row,
     )
 elif SQLA_VERSION < version.parse("2.0.0b1"):
     from .sqla14 import (
@@ -35,6 +59,7 @@ elif SQLA_VERSION < version.parse("2.0.0b1"):
         orm_query_keys,
         orm_result_type,
         orm_to_selectable,
+        Row,
     )
 else:
     from .sqla20 import (
@@ -46,6 +71,7 @@ else:
         orm_query_keys,
         orm_result_type,
         orm_to_selectable,
+        Row,
     )
 
 __all__ = [
@@ -58,4 +84,5 @@ __all__ = [
     "orm_query_keys",
     "orm_result_type",
     "orm_to_selectable",
+    "Row",
 ]
